@@ -2,6 +2,7 @@
    See COPYRIGHT.TXT and LICENSE.TXT for copyright and license information
    ----------------------------------------------------------------------- */
 #include "plmbase_config.h"
+#include <fstream>
 #include <stdio.h>
 #include <stdlib.h>
 #include <time.h>
@@ -11,17 +12,63 @@
 #include <omp.h>
 #endif
 
-#include "plmbase.h"
-#include "plmsys.h"
-
 #include "bspline_interpolate.h"
 #include "bspline_xform.h"
 #include "direction_cosines.h"
+#include "file_util.h"
 #include "interpolate_macros.h"
+#include "logfile.h"
 #include "plm_math.h"
-#include "plm_path.h"
+#include "print_and_exit.h"
+#include "string_util.h"
+#include "volume_header.h"
 #include "volume_macros.h"
 
+Bspline_xform::Bspline_xform ()
+{
+    this->coeff = 0;
+
+    this->lut_type = LUT_ALIGNED;
+
+    this->cidx_lut = 0;
+    this->c_lut = 0;
+    this->qidx_lut = 0;
+    this->q_lut = 0;
+
+    this->bx_lut = 0;
+    this->by_lut = 0;
+    this->bz_lut = 0;
+
+    this->ux_lut = 0;
+    this->uy_lut = 0;
+    this->uz_lut = 0;
+}
+
+Bspline_xform::~Bspline_xform ()
+{
+    if (this->coeff) {
+        free (this->coeff);
+    }
+    if (this->q_lut) {
+        free (this->q_lut);
+    }
+    if (this->c_lut) {
+        free (this->c_lut);
+    }
+    if (this->bx_lut) {
+        free (this->bx_lut);
+    }
+    if (this->by_lut) {
+        free (this->by_lut);
+    }
+    if (this->bz_lut) {
+        free (this->bz_lut);
+    }
+    delete[] ux_lut;
+    delete[] uy_lut;
+    delete[] uz_lut;
+}
+
 static float
 bspline_basis_eval (
     int t_idx, 
@@ -51,79 +98,49 @@ bspline_basis_eval (
 }
 
 void
-bspline_xform_set_default (Bspline_xform* bxf)
-{
-    int d;
-
-    memset (bxf, 0, sizeof (Bspline_xform));
-
-    for (d = 0; d < 3; d++) {
-        bxf->img_origin[d] = 0.0f;
-        bxf->img_spacing[d] = 1.0f;
-        bxf->img_dim[d] = 0;
-        bxf->roi_offset[d] = 0;
-        bxf->roi_dim[d] = 0;
-        bxf->vox_per_rgn[d] = 30;
-        bxf->grid_spac[d] = 30.0f;
-    }
-}
-
-void
-bspline_xform_save (Bspline_xform* bxf, const char* filename)
+Bspline_xform::save (const char* filename)
 {
     FILE* fp;
 
-    make_directory_recursive (filename);
+    make_parent_directories (filename);
     fp = fopen (filename, "wb");
     if (!fp) return;
 
     fprintf (fp, "MGH_GPUIT_BSP <experimental>\n");
     fprintf (fp, "img_origin = %f %f %f\n", 
-        bxf->img_origin[0], bxf->img_origin[1], bxf->img_origin[2]);
+        this->img_origin[0], this->img_origin[1], this->img_origin[2]);
     fprintf (fp, "img_spacing = %f %f %f\n", 
-        bxf->img_spacing[0], bxf->img_spacing[1], bxf->img_spacing[2]);
+        this->img_spacing[0], this->img_spacing[1], this->img_spacing[2]);
     fprintf (fp, "img_dim = %u %u %u\n", 
-        (unsigned int) bxf->img_dim[0], (unsigned int) bxf->img_dim[1], 
-        (unsigned int) bxf->img_dim[2]);
+        (unsigned int) this->img_dim[0], (unsigned int) this->img_dim[1], 
+        (unsigned int) this->img_dim[2]);
     fprintf (fp, "roi_offset = %d %d %d\n", 
-        (unsigned int) bxf->roi_offset[0], (unsigned int) bxf->roi_offset[1], 
-        (unsigned int) bxf->roi_offset[2]);
+        (unsigned int) this->roi_offset[0], (unsigned int) this->roi_offset[1], 
+        (unsigned int) this->roi_offset[2]);
     fprintf (fp, "roi_dim = %d %d %d\n", 
-        (unsigned int) bxf->roi_dim[0], (unsigned int) bxf->roi_dim[1], 
-        (unsigned int) bxf->roi_dim[2]);
+        (unsigned int) this->roi_dim[0], (unsigned int) this->roi_dim[1], 
+        (unsigned int) this->roi_dim[2]);
     fprintf (fp, "vox_per_rgn = %d %d %d\n", 
-        (unsigned int) bxf->vox_per_rgn[0], 
-        (unsigned int) bxf->vox_per_rgn[1], 
-        (unsigned int) bxf->vox_per_rgn[2]);
+        (unsigned int) this->vox_per_rgn[0], 
+        (unsigned int) this->vox_per_rgn[1], 
+        (unsigned int) this->vox_per_rgn[2]);
+    float *direction_cosines = this->dc.get_matrix ();
     fprintf (fp, "direction_cosines = %f %f %f %f %f %f %f %f %f\n", 
-        (bxf->dc).m_direction_cosines[0], 
-        (bxf->dc).m_direction_cosines[1], 
-        (bxf->dc).m_direction_cosines[2], 
-        (bxf->dc).m_direction_cosines[3], 
-        (bxf->dc).m_direction_cosines[4], 
-        (bxf->dc).m_direction_cosines[5], 
-        (bxf->dc).m_direction_cosines[6], 
-        (bxf->dc).m_direction_cosines[7], 
-        (bxf->dc).m_direction_cosines[8]);
+        direction_cosines[0], 
+        direction_cosines[1], 
+        direction_cosines[2], 
+        direction_cosines[3], 
+        direction_cosines[4], 
+        direction_cosines[5], 
+        direction_cosines[6], 
+        direction_cosines[7], 
+        direction_cosines[8]);
     /* No need to save grid_spac */
 
-#if defined (commentout)
-    {
-        /* This dumps in native, interleaved format */
-        for (i = 0; i < bxf->num_coeff; i++) {
-            fprintf (fp, "%6.3f\n", bxf->coeff[i]);
-        }
-    }
-#endif
-
     /* This dumps in itk-like planar format */
-    {
-        int i, j;
-        for (i = 0; i < 3; i++) {
-            for (j = 0; j < bxf->num_coeff / 3; j++) {
-                //fprintf (fp, "%6.3f\n", bxf->coeff[j*3 + i]);
-                fprintf (fp, "%.20f\n", bxf->coeff[j*3 + i]);
-            }
+    for (int i = 0; i < 3; i++) {
+        for (int j = 0; j < this->num_coeff / 3; j++) {
+            fprintf (fp, "%.20f\n", this->coeff[j*3 + i]);
         }
     }           
 
@@ -133,123 +150,135 @@ bspline_xform_save (Bspline_xform* bxf, const char* filename)
 Bspline_xform* 
 bspline_xform_load (const char* filename)
 {
-    Bspline_xform* bxf;
-    char buf[1024];
-    FILE* fp;
     int rc;
-    float img_origin[3];         /* Image origin (in mm) */
-    float img_spacing[3];        /* Image spacing (in mm) */
+    float img_origin[3] = {      /* Image origin (in mm) */
+        0., 0., 0. };
+    float img_spacing[3] = {     /* Image spacing (in mm) */
+        1., 1., 1. };
     unsigned int a, b, c;        /* For fscanf */
-    plm_long img_dim[3];           /* Image size (in vox) */
-    plm_long roi_offset[3];      /* Position of first vox in ROI (in vox) */
-    plm_long roi_dim[3];                 /* Dimension of ROI (in vox) */
-    plm_long vox_per_rgn[3];     /* Knot spacing (in vox) */
-    float dc[9];                 /* Direction cosines */
+    plm_long img_dim[3] = {      /* Image size (in vox) */
+        0, 0, 0 };
+    plm_long roi_offset[3] = {   /* Position of first vox in ROI (in vox) */
+        0, 0, 0 };
+    plm_long roi_dim[3] = {      /* Dimension of ROI (in vox) */
+        0, 0, 0 };
+    plm_long vox_per_rgn[3] = {  /* Knot spacing (in vox) */
+        0, 0, 0 };
+    float dc[9] = {              /* Direction cosines */
+        1., 0., 0., 0., 1., 0., 0., 0., 1. };
 
-    fp = fopen (filename, "r");
-    if (!fp) return 0;
-
-    /* Initialize parms */
-    bxf = (Bspline_xform*) malloc (sizeof(Bspline_xform));
-    bspline_xform_set_default (bxf);
-
-    /* Skip first line */
-    if (fgets (buf, 1024, fp) == NULL) {
-        logfile_printf ("File error.\n");
-        goto free_exit;
+    std::ifstream ifs (filename);
+    if (!ifs.is_open()) {
+        return 0;
     }
 
+    /* Check magic number */
+    std::string line;
+    getline (ifs, line);
+    if (!string_starts_with (line, "MGH_GPUIT_BSP")) {
+        return 0;
+    }
 
-    /* Read header */
-    rc = fscanf (fp, "img_origin = %f %f %f\n", 
-        &img_origin[0], &img_origin[1], &img_origin[2]);
-    if (rc != 3) {
-        logfile_printf ("Error parsing input xform (img_origin): %s\n", filename);
-        goto free_exit;
-    }
-    rc = fscanf (fp, "img_spacing = %f %f %f\n", 
-        &img_spacing[0], &img_spacing[1], &img_spacing[2]);
-    if (rc != 3) {
-        logfile_printf ("Error parsing input xform (img_spacing): %s\n", filename);
-        goto free_exit;
-    }
-    rc = fscanf (fp, "img_dim = %d %d %d\n", &a, &b, &c);
-    if (rc != 3) {
-        logfile_printf ("Error parsing input xform (img_dim): %s\n", filename);
-        goto free_exit;
-    }
-    img_dim[0] = a;
-    img_dim[1] = b;
-    img_dim[2] = c;
+    /* Parse header */
+    while (true) {
+        getline (ifs, line);
+        if (!ifs.good()) {
+            logfile_printf ("Error parsing bspline xform\n");
+            return 0;
+        }
 
-    rc = fscanf (fp, "roi_offset = %d %d %d\n", &a, &b, &c);
-    if (rc != 3) {
-        logfile_printf ("Error parsing input xform (roi_offset): %s\n", filename);
-        goto free_exit;
-    }
-    roi_offset[0] = a;
-    roi_offset[1] = b;
-    roi_offset[2] = c;
+        if (line.find('=') == std::string::npos) {
+            /* No "=" found.  Better be the first coefficient. */
+            break;
+        }
+        
+        rc = sscanf (line.c_str(), "img_origin = %f %f %f\n", 
+            &img_origin[0], &img_origin[1], &img_origin[2]);
+        if (rc == 3) continue;
 
-    rc = fscanf (fp, "roi_dim = %d %d %d\n", &a, &b, &c);
-    if (rc != 3) {
-        logfile_printf ("Error parsing input xform (roi_dim): %s\n", filename);
-        goto free_exit;
-    }
-    roi_dim[0] = a;
-    roi_dim[1] = b;
-    roi_dim[2] = c;
+        rc = sscanf (line.c_str(), "img_spacing = %f %f %f\n", 
+            &img_spacing[0], &img_spacing[1], &img_spacing[2]);
+        if (rc == 3) continue;
 
-    rc = fscanf (fp, "vox_per_rgn = %d %d %d\n", &a, &b, &c);
-    if (rc != 3) {
-        logfile_printf ("Error parsing input xform (vox_per_rgn): %s\n", filename);
-        goto free_exit;
-    }
-    vox_per_rgn[0] = a;
-    vox_per_rgn[1] = b;
-    vox_per_rgn[2] = c;
+        rc = sscanf (line.c_str(), "img_dim = %d %d %d\n", &a, &b, &c);
+        if (rc == 3) {
+            img_dim[0] = a;
+            img_dim[1] = b;
+            img_dim[2] = c;
+            continue;
+        }
 
-    /* JAS 2012.03.29 : check for direction cosines
-     * we must be careful because older plastimatch xforms do not have this */
-    rc = fscanf (fp, "direction_cosines = %f %f %f %f %f %f %f %f %f\n",
+        rc = sscanf (line.c_str(), "roi_offset = %d %d %d\n", &a, &b, &c);
+        if (rc == 3) {
+            roi_offset[0] = a;
+            roi_offset[1] = b;
+            roi_offset[2] = c;
+            continue;
+        }
+
+        rc = sscanf (line.c_str(), "roi_dim = %d %d %d\n", &a, &b, &c);
+        if (rc == 3) {
+            roi_dim[0] = a;
+            roi_dim[1] = b;
+            roi_dim[2] = c;
+            continue;
+        }
+
+        rc = sscanf (line.c_str(), "vox_per_rgn = %d %d %d\n", &a, &b, &c);
+        if (rc == 3) {
+            vox_per_rgn[0] = a;
+            vox_per_rgn[1] = b;
+            vox_per_rgn[2] = c;
+            continue;
+        }
+
+        rc = sscanf (line.c_str(), "direction_cosines = %f %f %f %f %f %f %f %f %f\n",
             &dc[0], &dc[1], &dc[2], &dc[3], &dc[4],
             &dc[5], &dc[6], &dc[7], &dc[8]);
-    if (rc != 9) {
-        dc[0] = 1.; dc[3] = 0.; dc[6] = 0.;
-        dc[1] = 0.; dc[4] = 1.; dc[7] = 0.;
-        dc[2] = 0.; dc[5] = 0.; dc[8] = 1.;
+        if (rc == 9) continue;
+
+        logfile_printf ("Error loading bxf file\n%s\n", line.c_str());
+        return 0;
     }
-    
+
+    logfile_printf ("1\n");
 
     /* Allocate memory and build LUTs */
-    bspline_xform_initialize (bxf, img_origin, img_spacing, img_dim,
+    Bspline_xform* bxf = new Bspline_xform;
+    bxf->initialize (img_origin, img_spacing, img_dim,
         roi_offset, roi_dim, vox_per_rgn, dc);
 
-    /* This loads from itk-like planar format */
-    {
-        int i, j;
-        for (i = 0; i < 3; i++) {
-            for (j = 0; j < bxf->num_coeff / 3; j++) {
-                rc = fscanf (fp, "%f\n", &bxf->coeff[j*3 + i]);
-                if (rc != 1) {
-                    logfile_printf ("Error parsing input xform (idx = %d,%d): %s\n", i, j, filename);
-                    bspline_xform_free (bxf);
-                    goto free_exit;
-                }
+    if (bxf->num_coeff < 1) {
+        logfile_printf ("Error loading bxf file, no coefficients\n");
+        delete bxf;
+        return 0;
+    }
+
+    /* Load from itk-like planar format */
+    for (int i = 0; i < 3; i++) {
+        for (int j = 0; j < bxf->num_coeff / 3; j++) {
+            /* The first line is already loaded from before */
+            if (i != 0 || j != 0) {
+                getline (ifs, line);
+            }
+            if (!ifs.good()) {
+                logfile_printf ("Error parsing bspline xform (idx = %d,%d): %s\n", i, j, filename);
+                delete bxf;
+                return 0;
+            }
+            rc = sscanf (line.c_str(), "%f", &bxf->coeff[j*3 + i]);
+            if (rc != 1) {
+                logfile_printf ("Error parsing bspline xform (idx = %d,%d): %s\n", i, j, filename);
+                delete bxf;
+                return 0;
             }
         }
     }
 
-    fclose (fp);
+    ifs.close();
     return bxf;
-
-  free_exit:
-    fclose (fp);
-    free (bxf);
-    return 0;
 }
 
-
 /* -----------------------------------------------------------------------
    Debugging routines
    ----------------------------------------------------------------------- */
@@ -331,78 +360,33 @@ bspline_xform_dump_luts (Bspline_xform* bxf)
 }
 
 void
-bspline_xform_set_coefficients (Bspline_xform* bxf, float val)
-{
-    int i;
-
-    for (i = 0; i < bxf->num_coeff; i++) {
-        bxf->coeff[i] = val;
-    }
-}
-
-void
-bspline_xform_initialize 
-(
-    Bspline_xform* bxf,           /* Output: bxf is initialized */
-    float img_origin[3],          /* Image origin (in mm) */
-    float img_spacing[3],         /* Image spacing (in mm) */
-    plm_long img_dim[3],          /* Image size (in vox) */
-    plm_long roi_offset[3],       /* Position of first vox in ROI (in vox) */
-    plm_long roi_dim[3],          /* Dimension of ROI (in vox) */
-    plm_long vox_per_rgn[3],      /* Knot spacing (in vox) */
-    float direction_cosines[9]    /* Direction cosines */
-)
+Bspline_xform::allocate ()
 {
     plm_long d;
     plm_long i, j, k, p;
     plm_long tx, ty, tz;
     float *A, *B, *C;
 
-    logfile_printf ("bspline_xform_initialize\n");
-    for (d = 0; d < 3; d++) {
-        /* copy input parameters over */
-        bxf->img_origin[d] = img_origin[d];
-        bxf->img_spacing[d] = img_spacing[d];
-        bxf->img_dim[d] = img_dim[d];
-        bxf->roi_offset[d] = roi_offset[d];
-        bxf->roi_dim[d] = roi_dim[d];
-        bxf->vox_per_rgn[d] = vox_per_rgn[d];
-        bxf->dc.set (direction_cosines);
-
-        /* grid spacing is in mm */
-        bxf->grid_spac[d] = bxf->vox_per_rgn[d] * fabs (bxf->img_spacing[d]);
-
-        /* rdims is the number of regions */
-        bxf->rdims[d] = 1 + (bxf->roi_dim[d] - 1) / bxf->vox_per_rgn[d];
-
-        /* cdims is the number of control points */
-        bxf->cdims[d] = 3 + bxf->rdims[d];
-    }
-
-    /* total number of control points & coefficients */
-    bxf->num_knots = bxf->cdims[0] * bxf->cdims[1] * bxf->cdims[2];
-    bxf->num_coeff = bxf->cdims[0] * bxf->cdims[1] * bxf->cdims[2] * 3;
-
     /* Allocate coefficients */
-    bxf->coeff = (float*) malloc (sizeof(float) * bxf->num_coeff);
-    memset (bxf->coeff, 0, sizeof(float) * bxf->num_coeff);
+    this->coeff = (float*) malloc (sizeof(float) * this->num_coeff);
+    memset (this->coeff, 0, sizeof(float) * this->num_coeff);
 
     /* Create q_lut */
-    bxf->q_lut = (float*) malloc (sizeof(float) 
-        * bxf->vox_per_rgn[0] 
-        * bxf->vox_per_rgn[1] 
-        * bxf->vox_per_rgn[2] 
+    this->q_lut = (float*) malloc (sizeof(float) 
+        * this->vox_per_rgn[0] 
+        * this->vox_per_rgn[1] 
+        * this->vox_per_rgn[2] 
         * 64);
-    if (!bxf->q_lut) {
+    if (!this->q_lut) {
         print_and_exit ("Error allocating memory for q_lut\n");
     }
 
-    A = (float*) malloc (sizeof(float) * bxf->vox_per_rgn[0] * 4);
-    B = (float*) malloc (sizeof(float) * bxf->vox_per_rgn[1] * 4);
-    C = (float*) malloc (sizeof(float) * bxf->vox_per_rgn[2] * 4);
+    A = (float*) malloc (sizeof(float) * this->vox_per_rgn[0] * 4);
+    B = (float*) malloc (sizeof(float) * this->vox_per_rgn[1] * 4);
+    C = (float*) malloc (sizeof(float) * this->vox_per_rgn[2] * 4);
 
-    for (i = 0; i < bxf->vox_per_rgn[0]; i++) {
-        float ii = ((float) i) / bxf->vox_per_rgn[0];
+    for (i = 0; i < this->vox_per_rgn[0]; i++) {
+        float ii = ((float) i) / this->vox_per_rgn[0];
         float t3 = ii*ii*ii;
         float t2 = ii*ii;
         float t1 = ii;
@@ -412,8 +396,8 @@ bspline_xform_initialize
         A[i*4+3] = (1.0/6.0) * (+ 1.0 * t3);
     }
 
-    for (j = 0; j < bxf->vox_per_rgn[1]; j++) {
-        float jj = ((float) j) / bxf->vox_per_rgn[1];
+    for (j = 0; j < this->vox_per_rgn[1]; j++) {
+        float jj = ((float) j) / this->vox_per_rgn[1];
         float t3 = jj*jj*jj;
         float t2 = jj*jj;
         float t1 = jj;
@@ -423,8 +407,8 @@ bspline_xform_initialize
         B[j*4+3] = (1.0/6.0) * (+ 1.0 * t3);
     }
 
-    for (k = 0; k < bxf->vox_per_rgn[2]; k++) {
-        float kk = ((float) k) / bxf->vox_per_rgn[2];
+    for (k = 0; k < this->vox_per_rgn[2]; k++) {
+        float kk = ((float) k) / this->vox_per_rgn[2];
         float t3 = kk*kk*kk;
         float t2 = kk*kk;
         float t1 = kk;
@@ -435,13 +419,13 @@ bspline_xform_initialize
     }
 
     p = 0;
-    for (k = 0; k < bxf->vox_per_rgn[2]; k++) {
-        for (j = 0; j < bxf->vox_per_rgn[1]; j++) {
-            for (i = 0; i < bxf->vox_per_rgn[0]; i++) {
+    for (k = 0; k < this->vox_per_rgn[2]; k++) {
+        for (j = 0; j < this->vox_per_rgn[1]; j++) {
+            for (i = 0; i < this->vox_per_rgn[0]; i++) {
                 for (tz = 0; tz < 4; tz++) {
                     for (ty = 0; ty < 4; ty++) {
                         for (tx = 0; tx < 4; tx++) {
-                            bxf->q_lut[p++] = A[i*4+tx] * B[j*4+ty] * C[k*4+tz];
+                            this->q_lut[p++] = A[i*4+tx] * B[j*4+ty] * C[k*4+tz];
                         }
                     }
                 }
@@ -453,21 +437,21 @@ bspline_xform_initialize
     free (A);
         
     /* Create c_lut */
-    bxf->c_lut = (plm_long*) malloc (sizeof(plm_long) 
-        * bxf->rdims[0] 
-        * bxf->rdims[1] 
-        * bxf->rdims[2] 
+    this->c_lut = (plm_long*) malloc (sizeof(plm_long) 
+        * this->rdims[0] 
+        * this->rdims[1] 
+        * this->rdims[2] 
         * 64);
     p = 0;
-    for (k = 0; k < bxf->rdims[2]; k++) {
-        for (j = 0; j < bxf->rdims[1]; j++) {
-            for (i = 0; i < bxf->rdims[0]; i++) {
+    for (k = 0; k < this->rdims[2]; k++) {
+        for (j = 0; j < this->rdims[1]; j++) {
+            for (i = 0; i < this->rdims[0]; i++) {
                 for (tz = 0; tz < 4; tz++) {
                     for (ty = 0; ty < 4; ty++) {
                         for (tx = 0; tx < 4; tx++) {
-                            bxf->c_lut[p++] = 
-                                + (k + tz) * bxf->cdims[0] * bxf->cdims[1]
-                                + (j + ty) * bxf->cdims[0] 
+                            this->c_lut[p++] = 
+                                + (k + tz) * this->cdims[0] * this->cdims[1]
+                                + (j + ty) * this->cdims[0] 
                                 + (i + tx);
                         }
                     }
@@ -477,30 +461,69 @@ bspline_xform_initialize
     }
 
     /* Create b_luts */
-    bxf->bx_lut = (float*)malloc(4*bxf->vox_per_rgn[0]*sizeof(float));
-    bxf->by_lut = (float*)malloc(4*bxf->vox_per_rgn[1]*sizeof(float));
-    bxf->bz_lut = (float*)malloc(4*bxf->vox_per_rgn[2]*sizeof(float));
+    this->bx_lut = (float*)malloc(4*this->vox_per_rgn[0]*sizeof(float));
+    this->by_lut = (float*)malloc(4*this->vox_per_rgn[1]*sizeof(float));
+    this->bz_lut = (float*)malloc(4*this->vox_per_rgn[2]*sizeof(float));
 
     for (int j=0; j<4; j++) {
-        for (int i=0; i<bxf->vox_per_rgn[0]; i++) {
-            bxf->bx_lut[i*4+j] = bspline_basis_eval (j, i, bxf->vox_per_rgn[0]);
+        for (int i=0; i<this->vox_per_rgn[0]; i++) {
+            this->bx_lut[i*4+j] = bspline_basis_eval (
+                j, i, this->vox_per_rgn[0]);
         }
-        for (int i=0; i<bxf->vox_per_rgn[1]; i++) {
-            bxf->by_lut[i*4+j] = bspline_basis_eval (j, i, bxf->vox_per_rgn[1]);
+        for (int i=0; i<this->vox_per_rgn[1]; i++) {
+            this->by_lut[i*4+j] = bspline_basis_eval (
+                j, i, this->vox_per_rgn[1]);
         }
-        for (int i=0; i<bxf->vox_per_rgn[2]; i++) {
-            bxf->bz_lut[i*4+j] = bspline_basis_eval (j, i, bxf->vox_per_rgn[2]);
+        for (int i=0; i<this->vox_per_rgn[2]; i++) {
+            this->bz_lut[i*4+j] = bspline_basis_eval (
+                j, i, this->vox_per_rgn[2]);
         }
     }
 
-    //dump_luts (bxf);
-
     logfile_printf ("rdims = (%d,%d,%d)\n", 
-        bxf->rdims[0], bxf->rdims[1], bxf->rdims[2]);
+        this->rdims[0], this->rdims[1], this->rdims[2]);
     logfile_printf ("vox_per_rgn = (%d,%d,%d)\n", 
-        bxf->vox_per_rgn[0], bxf->vox_per_rgn[1], bxf->vox_per_rgn[2]);
+        this->vox_per_rgn[0], this->vox_per_rgn[1], this->vox_per_rgn[2]);
     logfile_printf ("cdims = (%d %d %d)\n", 
-        bxf->cdims[0], bxf->cdims[1], bxf->cdims[2]);
+        this->cdims[0], this->cdims[1], this->cdims[2]);
+}
+
+void
+Bspline_xform::initialize 
+(
+    float img_origin[3],          /* Image origin (in mm) */
+    float img_spacing[3],         /* Image spacing (in mm) */
+    plm_long img_dim[3],          /* Image size (in vox) */
+    plm_long roi_offset[3],       /* Position of first vox in ROI (in vox) */
+    plm_long roi_dim[3],          /* Dimension of ROI (in vox) */
+    plm_long vox_per_rgn[3],      /* Knot spacing (in vox) */
+    float direction_cosines[9]    /* Direction cosines */
+)
+{
+    logfile_printf ("bspline_xform_initialize\n");
+
+    /* Initialize base class members */
+    this->set (img_origin, img_spacing, img_dim, roi_offset,
+        roi_dim, vox_per_rgn, direction_cosines);
+
+    /* Allocate and initialize coefficients and LUTs */
+    this->allocate ();
+}
+
+void
+Bspline_xform::initialize 
+(
+    const Plm_image_header *pih,
+    const float grid_spac[3]
+)
+{
+    logfile_printf ("bspline_xform_initialize\n");
+
+    /* Initialize base class members */
+    this->set (pih, grid_spac);
+
+    /* Allocate and initialize coefficients and LUTs */
+    this->allocate ();
 }
 
 /* -----------------------------------------------------------------------
@@ -596,25 +619,49 @@ bspline_xform_extend (
 }
 
 void
-bspline_xform_free (Bspline_xform* bxf)
+Bspline_xform::fill_coefficients (float val)
 {
-    free (bxf->coeff);
-    free (bxf->q_lut);
-    free (bxf->c_lut);
-    free (bxf->bx_lut);
-    free (bxf->by_lut);
-    free (bxf->bz_lut);
+    int i;
+
+    for (i = 0; i < this->num_coeff; i++) {
+        this->coeff[i] = val;
+    }
+}
+
+void
+Bspline_xform::jitter_if_zero ()
+{
+    /*   The MI algorithm will get stuck for a set of coefficients all equaling
+     *   zero due to the method we use to compute the cost function gradient.
+     *   However, it is possible we could be inheriting coefficients from a
+     *   prior stage, so we must check for inherited coefficients before
+     *   applying an initial offset to the coefficient array. */
+    for (int i = 0; i < this->num_coeff; i++) {
+        if (this->coeff[i] != 0.0f) {
+            return;
+        }
+    }
+    fill_coefficients (0.01f);
 }
 
 /* Set volume header from B-spline Xform */
 void 
 Bspline_xform::get_volume_header (Volume_header *vh)
 {
-#if 0
-    vh->set_dim (this->img_dim);
-    vh->set_origin (this->img_origin);
-    vh->set_spacing (this->img_spacing);
-#endif
-
-    vh->set (this->img_dim, this->img_origin, this->img_spacing, (this->dc).m_direction_cosines);
+    vh->set (this->img_dim, this->img_origin, this->img_spacing, 
+        this->dc.get_matrix());
 }
+
+void
+Bspline_xform::log_header ()
+{
+    logfile_printf ("BSPLINE XFORM HEADER\n");
+    logfile_printf ("vox_per_rgn = %d %d %d\n", 
+        this->vox_per_rgn[0], this->vox_per_rgn[1], this->vox_per_rgn[2]);
+    logfile_printf ("roi_offset = %d %d %d\n", 
+        this->roi_offset[0], this->roi_offset[1], this->roi_offset[2]);
+    logfile_printf ("roi_dim = %d %d %d\n", 
+        this->roi_dim[0], this->roi_dim[1], this->roi_dim[2]);
+    logfile_printf ("img_dc = %s\n", this->dc.get_string().c_str());
+}
+
